@@ -12,8 +12,11 @@
 # include <pwd.h>
 # include <grp.h>
 # include <dirent.h>
+# include <sys/time.h>
+# include <sys/utsname.h>
 #endif
 #ifdef RKTIO_SYSTEM_WINDOWS
+# include <windows.h>
 # include <shlobj.h>
 # include <direct.h>
 # include <sys/stat.h>
@@ -129,12 +132,12 @@ static time_t convert_date(const FILETIME *ft)
   SYSTEMTIME st;
   TIME_ZONE_INFORMATION tz;
 
-  /* FindFirstFile incorrectly shifts for daylight saving. It
+  /* GetFileAttributes incorrectly shifts for daylight saving. It
      subtracts an hour to get to UTC when daylight saving is in effect
      now, even when daylight saving was not in effect when the file
      was saved.  Counteract the difference. There's a race condition
      here, because we might cross the daylight-saving boundary between
-     the time that FindFirstFile runs and GetTimeZoneInformation
+     the time that GetFileAttributes runs and GetTimeZoneInformation
      runs. Cross your fingers... */
   FileTimeToLocalFileTime(ft, &ft2);
   FileTimeToSystemTime(&ft2, &st);
@@ -143,47 +146,8 @@ static time_t convert_date(const FILETIME *ft)
   if (GetTimeZoneInformation(&tz) == TIME_ZONE_ID_DAYLIGHT) {
     /* Daylight saving is in effect now, so there may be a bad
        shift. Check the file's date. */
-    int start_day_of_month, end_day_of_month, first_day_of_week, diff, end_shift;
-
-    /* Valid only when the months match: */
-    first_day_of_week = (st.wDayOfWeek - (st.wDay - 1 - (((st.wDay - 1) / 7) * 7)));
-    if (first_day_of_week < 0)
-      first_day_of_week += 7;
-
-    diff = (tz.DaylightDate.wDayOfWeek - first_day_of_week);
-    if (diff < 0)
-      diff += 7;
-    start_day_of_month = 1 + (((tz.DaylightDate.wDay - 1) * 7)
-			      + diff);
-	
-    diff = (tz.StandardDate.wDayOfWeek - first_day_of_week);
-    if (diff < 0)
-      diff += 7;
-    end_day_of_month = 1 + (((tz.StandardDate.wDay - 1) * 7)
-			    + diff);
-
-    /* Count ambigious range (when the clock goes back) as
-       in standard time. We assume that subtracting the 
-       ambiguous range does not go back into the previous day,
-       and that the shift is a multiple of an hour. */
-    end_shift = ((tz.StandardBias - tz.DaylightBias) / 60);
-
-    if ((st.wMonth < tz.DaylightDate.wMonth)
-	|| ((st.wMonth == tz.DaylightDate.wMonth)
-	    && ((st.wDay < start_day_of_month)
-		|| ((st.wDay == start_day_of_month)
-		    && (st.wHour < tz.DaylightDate.wHour))))) {
-      /* Daylight saving had not yet started. */
+    if (!rktio_system_time_is_dst(&st, NULL))
       delta = ((tz.StandardBias - tz.DaylightBias) * 60);
-    } else if ((st.wMonth > tz.StandardDate.wMonth)
-	       || ((st.wMonth == tz.StandardDate.wMonth)
-		   && ((st.wDay > end_day_of_month)
-		       || ((st.wDay == end_day_of_month)
-			   && (st.wHour >= (tz.StandardDate.wHour
-					    - end_shift)))))) {
-      /* Daylight saving was already over. */
-      delta = ((tz.StandardBias - tz.DaylightBias) * 60);
-    }
   }
 
   l = ((((LONGLONG)ft->dwHighDateTime << 32) | ft->dwLowDateTime)
@@ -349,6 +313,7 @@ static int UNC_stat(rktio_t *rktio, const char *dirname, int *flags, int *isdir,
   char *copy;
   WIN32_FILE_ATTRIBUTE_DATA fad;
   int len, must_be_dir = 0;
+  int same_path = 0; /* to give up on cyclic links */
   const wchar_t *wp;
 
   if (resolved_path)
@@ -410,7 +375,7 @@ static int UNC_stat(rktio_t *rktio, const char *dirname, int *flags, int *isdir,
 
   wp = WIDE_PATH_temp(copy);
   if (!wp) {
-    /* Treat invalid path as non-existent */
+    /* Treat invalid path as non-existent; `WIDE_PATH_temp` set the error */
     free(copy);
     return 0;
   }
@@ -420,9 +385,11 @@ static int UNC_stat(rktio_t *rktio, const char *dirname, int *flags, int *isdir,
     free(copy);
     return 0;
   } else {
-    if (GET_FF_ATTRIBS(fad) & FF_A_LINK) {
+    if ((GET_FF_ATTRIBS(fad) & FF_A_LINK) && !same_path) {
       if (islink) {
 	*islink = 1;
+        if (isdir)
+          *isdir = (GET_FF_ATTRIBS(fad) & FF_A_DIR);
 	return 1;
       } else {
 	/* Resolve a link by opening the link and then getting
@@ -444,7 +411,7 @@ static int UNC_stat(rktio_t *rktio, const char *dirname, int *flags, int *isdir,
 	}
 
         do {
-	  init_procs();
+          init_procs();
           if (dest) free(dest);
           dest_len = len + 1;
           dest = malloc(dest_len);
@@ -465,6 +432,8 @@ static int UNC_stat(rktio_t *rktio, const char *dirname, int *flags, int *isdir,
 	if (resolved_path)
 	  *resolved_path = dirname;
 
+	same_path = !strcmp(dirname, copy);
+
         free(dest);
         free(copy);
 
@@ -623,6 +592,46 @@ int rktio_link_exists(rktio_t *rktio, const char *filename)
 #endif
 }
 
+int rktio_file_type(rktio_t *rktio, rktio_const_string_t filename)
+/* Windows: check for special filenames before calling */
+{
+#ifdef RKTIO_SYSTEM_WINDOWS
+  {
+    int islink, isdir;
+    if (UNC_stat(rktio, filename, NULL, &isdir, &islink, NULL, NULL, NULL, -1)) {
+      if (islink) {
+        if (isdir)
+          return RKTIO_FILE_TYPE_DIRECTORY_LINK;
+        else
+          return RKTIO_FILE_TYPE_LINK;
+      } else if (isdir)
+        return RKTIO_FILE_TYPE_DIRECTORY;
+      else
+        return RKTIO_FILE_TYPE_FILE;
+    } else
+      return RKTIO_FILE_TYPE_ERROR;
+  }
+#else
+  {
+    struct MSC_IZE(stat) buf;
+    while (1) {
+      if (!MSC_W_IZE(lstat)(MSC_WIDE_PATH_temp(filename), &buf))
+	break;
+      else if (errno != EINTR)
+	return RKTIO_FILE_TYPE_ERROR;
+    }
+
+    if (S_ISLNK(buf.st_mode))
+      return RKTIO_FILE_TYPE_LINK;
+    else if (S_ISDIR(buf.st_mode))
+      return RKTIO_FILE_TYPE_DIRECTORY;
+    else
+      return RKTIO_FILE_TYPE_FILE;
+  }
+#endif
+  
+}
+
 char *rktio_get_current_directory(rktio_t *rktio)
 {
 #ifdef RKTIO_SYSTEM_WINDOWS
@@ -630,7 +639,7 @@ char *rktio_get_current_directory(rktio_t *rktio)
  wchar_t *wbuf;
  char *r;
 
- wbuf = malloc(bl);
+ wbuf = malloc(bl * sizeof(wchar_t));
  while (1) {
    need_l = GetCurrentDirectoryW(bl, wbuf);
    if (need_l > bl) {
@@ -665,8 +674,10 @@ char *rktio_get_current_directory(rktio_t *rktio)
    } else
      break;
  }
- if (!r)
+ if (!r) {
+   free(s);
    get_posix_error();
+ }
  return r;
 #endif
 }
@@ -735,7 +746,7 @@ static rktio_identity_t *get_identity(rktio_t *rktio, rktio_fd_t *fd, const char
     if (!wp) return 0;
     fdh = CreateFileW(wp,
                       0, /* not even read access => just get info */
-                      FILE_SHARE_READ | FILE_SHARE_WRITE,
+                      FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS 
@@ -923,6 +934,7 @@ char *rktio_readlink(rktio_t *rktio, const char *fullfilename)
           set_racket_error(RKTIO_ERROR_NOT_A_LINK);
         else
           get_posix_error();
+        free(buffer);
         return NULL;
       }
     } else if (len == buf_len) {
@@ -951,7 +963,8 @@ int rktio_make_directory(rktio_t *rktio, const char *filename)
   /* Make sure path doesn't have trailing separator: */
   len = strlen(filename);
   while (len && IS_A_SEP(filename[len - 1])) {
-    copied = MSC_IZE(strdup)(filename);
+    if (!copied)
+      copied = MSC_IZE(strdup)(filename);
     copied[--len] = 0;
     filename = copied;
   }
@@ -1018,16 +1031,21 @@ int rktio_make_link(rktio_t *rktio, const char *src, const char *dest, int dest_
 {
 #if defined(RKTIO_SYSTEM_WINDOWS)
   init_procs();
-    
+
+# ifndef SYMBOLIC_LINK_FLAG_DIRECTORY
+#  define SYMBOLIC_LINK_FLAG_DIRECTORY 0x1
+# endif
+# ifndef SYMBOLIC_LINK_FLAG_ALLOW_UNPRIVILEGED_CREATE
+#  define SYMBOLIC_LINK_FLAG_ALLOW_UNPRIVILEGED_CREATE 0x2
+# endif
+
   if (CreateSymbolicLinkProc) {
-    int flags;
+    int flags = SYMBOLIC_LINK_FLAG_ALLOW_UNPRIVILEGED_CREATE;
     wchar_t *src_w;
     wchar_t *dest_w;
 
     if (dest_is_directory)
-      flags = 0x1; /* directory */
-    else
-      flags = 0; /* file */
+      flags |= SYMBOLIC_LINK_FLAG_DIRECTORY; /* directory */
 
     src_w = WIDE_PATH_copy(src);
     if (!src_w) return 0;
@@ -1519,12 +1537,13 @@ char *rktio_directory_list_step(rktio_t *rktio, rktio_directory_list_t *dl)
   while ((e = readdir(dl->dir))) {
     int nlen;
 
-# ifdef DIRENT_NO_NAMLEN
-    nlen = strlen(e->d_name);
-# elif defined(__QNX__) || defined(__QNXNTO__)
+# ifdef HAVE_DIRENT_NAMLEN
+    nlen = e->d_namlen;
+# elif HAVE_DIRENT_NAMELEN
+    /* Case for QNX - which seems to define d_namelen instead */
     nlen = e->d_namelen;
 # else
-    nlen = e->d_namlen;
+    nlen = strlen(e->d_name);
 # endif
 
 # if defined(RKTIO_SYSTEM_UNIX) || defined(RKTIO_SYSTEM_WINDOWS)
@@ -1898,6 +1917,20 @@ static char *append_paths(char *a, char *b, int free_a, int free_b)
   return s;
 }
 
+#ifdef RKTIO_SYSTEM_UNIX
+static int directory_or_file_exists(rktio_t *rktio, char *dir, char *maybe_file)
+{
+  if (maybe_file) {
+    int r;
+    char *path = append_paths(dir, maybe_file, 0, 0);
+    r = rktio_file_exists(rktio, path);
+    free(path);
+    return r;
+  } else
+    return rktio_directory_exists(rktio, dir);
+}
+#endif
+
 char *rktio_system_path(rktio_t *rktio, int which)
 {
 #ifdef RKTIO_SYSTEM_UNIX
@@ -1929,21 +1962,62 @@ char *rktio_system_path(rktio_t *rktio, int which)
   
   {
     /* Everything else is in ~: */
-    char *home_str, *alt_home, *home;
+    char *home_str, *alt_home, *home, *default_xdg_home_str = NULL, *xdg_home_str = NULL, *prefer_home;
+    char *home_file = NULL, *prefer_home_file = NULL;
+
+    alt_home = rktio_getenv(rktio, "PLTUSERHOME");
 
     if ((which == RKTIO_PATH_PREF_DIR) 
-	|| (which == RKTIO_PATH_PREF_FILE)
-	|| (which == RKTIO_PATH_ADDON_DIR)) {
-#if defined(OS_X) && !defined(XONX)
+        || (which == RKTIO_PATH_PREF_FILE)
+        || (which == RKTIO_PATH_ADDON_DIR)
+        || (which == RKTIO_PATH_CACHE_DIR)
+        || (which == RKTIO_PATH_INIT_DIR)
+        || (which == RKTIO_PATH_INIT_FILE)) {
+#if defined(OS_X) && !defined(RACKET_XONX) && !defined(XONX)
       if (which == RKTIO_PATH_ADDON_DIR)
 	home_str = "~/Library/Racket/";
-      else
+      else if (which == RKTIO_PATH_CACHE_DIR)
+	home_str = "~/Library/Caches/Racket/";
+      else if ((which == RKTIO_PATH_INIT_DIR)
+               || (which == RKTIO_PATH_INIT_FILE)) {
+        default_xdg_home_str = "~/Library/Racket/";
+        prefer_home_file = "racketrc.rktl";
+        home_str = "~/";
+        home_file = ".racketrc";
+      } else
 	home_str = "~/Library/Preferences/";
 #else
-      home_str = "~/.racket/";
-#endif 
+      char *envvar, *xdg_dir;
+      if (which == RKTIO_PATH_ADDON_DIR) {
+        default_xdg_home_str = "~/.local/share/racket/";
+        envvar = "XDG_DATA_HOME";
+      } else if (which == RKTIO_PATH_CACHE_DIR) {
+        default_xdg_home_str = "~/.cache/racket/";
+        envvar = "XDG_CACHE_HOME";
+      } else {
+        default_xdg_home_str = "~/.config/racket/";
+        envvar = "XDG_CONFIG_HOME";
+      }
+      if (alt_home)
+        xdg_dir = NULL;
+      else
+        xdg_dir = rktio_getenv(rktio, envvar);
+      /* xdg_dir is invalid if it is not an absolute path */
+      if (xdg_dir && (strlen(xdg_dir) > 0) && (xdg_dir[0] == '/')) {
+        xdg_home_str = append_paths(xdg_dir, "racket/", 1, 0);
+      } else {
+        if (xdg_dir) free(xdg_dir);
+      }
+
+      if ((which == RKTIO_PATH_INIT_DIR) || (which == RKTIO_PATH_INIT_FILE)) {
+        home_str = "~/";
+        home_file = ".racketrc";
+      } else { /* RKTIO_PATH_{ADDON_DIR,PREF_DIR,PREF_FILE,CACHE_DIR} */
+        home_str = "~/.racket/";
+      }
+#endif
     } else {
-#if defined(OS_X) && !defined(XONX)
+#if defined(OS_X) && !defined(RACKET_XONX) && !defined(XONX)
       if (which == RKTIO_PATH_DESK_DIR)
 	home_str = "~/Desktop/";
       else if (which == RKTIO_PATH_DOC_DIR)
@@ -1953,35 +2027,69 @@ char *rktio_system_path(rktio_t *rktio, int which)
         home_str = "~/";
     }
 
-    alt_home = rktio_getenv(rktio, "PLTUSERHOME");
-    if (alt_home)
-      home = append_paths(alt_home, home_str + 2, 1, 0);
-    else {
-      home = rktio_expand_user_tilde(rktio, home_str);
-      
-      if (!home) {
-        /* Something went wrong with the user lookup. Just drop "~'. */
-        int h_len = strlen(home_str);
-        home = (char *)malloc(h_len - 2 + 1);
-        strcpy(home, home_str+2);
+    /* If `xdg_home_str` is non-NULL, it must be an absolute
+       path that is `malloc`ed.
+       If `default_xdg_home_str` is non-NULL, it starts with "~/"
+       and is not `malloc`ed. */
+
+    if (xdg_home_str || default_xdg_home_str) {
+      if (xdg_home_str)
+        prefer_home = xdg_home_str;
+      else if (alt_home)
+        prefer_home = append_paths(alt_home, default_xdg_home_str + 2, 0, 0);
+      else
+        prefer_home = rktio_expand_user_tilde(rktio, default_xdg_home_str);
+
+      if (directory_or_file_exists(rktio, prefer_home, prefer_home_file))
+        home_str = NULL;
+    } else
+      prefer_home = NULL;
+
+    if (home_str) {
+      if (alt_home)
+        home = append_paths(alt_home, home_str + 2, 1, 0);
+      else
+        home = rktio_expand_user_tilde(rktio, home_str);
+
+      if (prefer_home) {
+        if (!directory_or_file_exists(rktio, home, home_file)) {
+          free(home);
+          home = prefer_home;
+        } else {
+          free(prefer_home);
+          prefer_home = NULL;
+        }
       }
-    }
-    
+    } else
+      home = prefer_home;
+
+    /* At this point, we're using `home`, but `prefer_home` can still
+       be non-NULL and equal to `home` to mean that we should use
+       XDG-style file names. */
+
     if ((which == RKTIO_PATH_PREF_DIR) || (which == RKTIO_PATH_INIT_DIR) 
 	|| (which == RKTIO_PATH_HOME_DIR) || (which == RKTIO_PATH_ADDON_DIR)
-	|| (which == RKTIO_PATH_DESK_DIR) || (which == RKTIO_PATH_DOC_DIR))
+	|| (which == RKTIO_PATH_DESK_DIR) || (which == RKTIO_PATH_DOC_DIR)
+        || (which == RKTIO_PATH_CACHE_DIR))
       return home;
 
-    if (which == RKTIO_PATH_INIT_FILE)
-      return append_paths(home, ".racketrc", 1, 0);
+    if (which == RKTIO_PATH_INIT_FILE) {
+      if (prefer_home)
+        return append_paths(prefer_home, "racketrc.rktl", 1, 0);
+      else
+        return append_paths(home, ".racketrc", 1, 0);
+    }
+
     if (which == RKTIO_PATH_PREF_FILE) {
-#if defined(OS_X) && !defined(XONX)
+#if defined(OS_X) && !defined(RACKET_XONX) && !defined(XONX)
       return append_paths(home, "org.racket-lang.prefs.rktd", 1, 0);
-#else      
+#else
       return append_paths(home, "racket-prefs.rktd", 1, 0);
 #endif
-    } else
+    } else {
+      free(home);
       return strdup("/");
+    }
   }
 #endif
 
@@ -2023,6 +2131,7 @@ char *rktio_system_path(rktio_t *rktio, int which)
       int which_folder;
 
       if ((which == RKTIO_PATH_ADDON_DIR)
+          || (which == RKTIO_PATH_CACHE_DIR) /* maybe CSIDL_LOCAL_APPDATA instead? */
 	  || (which == RKTIO_PATH_PREF_DIR)
 	  || (which == RKTIO_PATH_PREF_FILE)) 
 	which_folder = CSIDL_APPDATA;
@@ -2116,6 +2225,7 @@ char *rktio_system_path(rktio_t *rktio, int which)
       return home;
 
     if ((which == RKTIO_PATH_ADDON_DIR)
+        || (which == RKTIO_PATH_CACHE_DIR)
 	|| (which == RKTIO_PATH_PREF_DIR)
 	|| (which == RKTIO_PATH_PREF_FILE)) {
       home = append_paths(home, "Racket", 1, 0);
@@ -2129,3 +2239,91 @@ char *rktio_system_path(rktio_t *rktio, int which)
   }
 #endif
 }
+
+/*========================================================================*/
+/* system information as a string                                         */
+/*========================================================================*/
+
+
+#ifdef RKTIO_SYSTEM_UNIX
+char *rktio_uname(rktio_t *rktio) {
+  char *s;
+  struct utsname u;
+  int ok, len;
+  int syslen, nodelen, rellen, verlen, machlen;
+
+  do {
+    ok = uname(&u);
+  } while ((ok == -1) && (errno == EINTR));
+    
+  if (ok != 0)
+    return strdup("<unknown machine>");
+
+  syslen = strlen(u.sysname);
+  nodelen = strlen(u.nodename);
+  rellen = strlen(u.release);
+  verlen = strlen(u.version);
+  machlen = strlen(u.machine);
+
+  len = (syslen + 1 + nodelen + 1 + rellen + 1 + verlen + 1 + machlen + 1);
+
+  s = malloc(len);
+
+# define ADD_UNAME_STR(sn, slen) do {                  \
+    memcpy(s + len, sn, slen);                         \
+    len += slen;                                       \
+    s[len++] = ' ';                                    \
+  } while (0)
+
+  len = 0;
+  ADD_UNAME_STR(u.sysname, syslen);
+  ADD_UNAME_STR(u.nodename, nodelen);
+  ADD_UNAME_STR(u.release, rellen);
+  ADD_UNAME_STR(u.version, verlen);
+  ADD_UNAME_STR(u.machine, machlen);
+  s[len - 1] = 0;
+
+# undef ADD_UNAME_STR
+  
+  return s;
+}
+#endif
+ 
+#ifdef RKTIO_SYSTEM_WINDOWS
+char *rktio_uname(rktio_t *rktio) {
+  char buff[1024];
+  OSVERSIONINFO info;
+  BOOL hasInfo;
+  char *p;
+
+  info.dwOSVersionInfoSize = sizeof(info);
+
+  GetVersionEx(&info);
+
+  hasInfo = FALSE;
+
+  p = info.szCSDVersion;
+
+  while (p < info.szCSDVersion + sizeof(info.szCSDVersion) &&
+	 *p) {
+    if (*p != ' ') {
+      hasInfo = TRUE;
+      break;
+    }
+    p = p + 1;
+  }
+
+  sprintf(buff,"Windows %s %ld.%ld (Build %ld)%s%s",
+	  (info.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) ?
+	  "9x" :
+	  (info.dwPlatformId == VER_PLATFORM_WIN32_NT) ?
+	  "NT" : "Unknown platform",
+	  info.dwMajorVersion,info.dwMinorVersion,
+	  (info.dwPlatformId == VER_PLATFORM_WIN32_NT) ?
+	  info.dwBuildNumber :
+	  info.dwBuildNumber & 0xFFFF,
+	  hasInfo ? " " : "",hasInfo ? info.szCSDVersion : "");
+
+  return strdup(buff);
+}
+#endif
